﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

public class MethodCallCache {
	// TODO: This class should be implemented as thread-safe.

	static List<List<Tuple<object[], object>>> cache = new List<List<Tuple<object[], object>>>();
	static Stack<Tuple<int, object[]>> callStack = new Stack<Tuple<int, object[]>>();

	static bool FindMethodCall(int methodId, object[] methodArgs, out object methodRetValue) {
		if (cache.Count <= methodId) {
			for (int i = 0; i < methodId - cache.Count + 1; i++) {
				cache.Add(new List<Tuple<object[], object>>());
			}

			methodRetValue = null;
			return false;
		}

		var methodInvocationList = cache[methodId];

		foreach (var invocation in methodInvocationList) {
			bool match = true;
			for (int i = 0; i < methodArgs.Length; i++) {
				if (!invocation.Item1[i].Equals(methodArgs[i])) {
					match = false;
					break;
				}
			}

			if (match) {
				methodRetValue = invocation.Item2;
				return true;
			}
		}

		methodRetValue = null;
		return false;
	}

	public static bool MethodBegin(int methodId, object[] methodArgs, out object methodRetValue) {
		if (FindMethodCall(methodId, methodArgs, out methodRetValue)) {
			Console.WriteLine("--- MethodBegin: CACHED call of method ID {0} invoked, returning {1}.", methodId, methodRetValue);
			return true;
		} else {
			Console.WriteLine("--- MethodBegin: NEW call of method ID {0} invoked.", methodId);
			callStack.Push(new Tuple<int, object[]>(methodId, methodArgs));
			return false;
		}
	}

	public static void MethodEnd(int methodId, object methodRetValue) {
		var methodInvocation = callStack.Pop();
		if (methodInvocation.Item1 != methodId) {
			throw new InvalidOperationException(string.Format("MethodEnd for {0} called, not for {1} as expected.", methodId, methodInvocation.Item1));
		}

		object oldRetValue;
		if (FindMethodCall(methodId, methodInvocation.Item2, out oldRetValue)) {
			// TODO: Check return value did not change.
		} else {
			cache[methodId].Add(new Tuple<object[], object>(methodInvocation.Item2, methodRetValue));
		}

		Console.WriteLine("--- MethodEnd: call of method ID {0} ending with {1}.", methodId, methodRetValue);
	}
}

